#include <stdio.h>
#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <gdk/gdkx.h>
#include <X11/Xlib.h>

static void on_destroy( GtkWidget *widget, gpointer data )
{
    gtk_main_quit();
}

enum
{
    ID_COLUMN,
    NAME_COLUMN,
    X_COLUMN,
    Y_COLUMN,
    W_COLUMN,
    H_COLUMN,
    EVENTS_COLUMN,
    OVERRIDE_COLUMN,
    N_COLUMNS
};

void print_event_mask( char *data, long mask )
{
    int seen = 0;
    if( mask & KeyPressMask || mask & KeyReleaseMask ||
        mask & KeymapStateMask) {
        if( seen ) { data += sprintf( data, ", " ); } else { seen = 1; }
        data += sprintf( data, "Key" );
    }
    if( mask & ButtonPressMask ||  mask & ButtonReleaseMask ) {
        if( seen ) { data += sprintf( data, ", " ); } else { seen = 1; }
        data += sprintf( data, "Button" );
    }
    if( mask & EnterWindowMask || mask & LeaveWindowMask ) {
        if( seen ) { data += sprintf( data, ", " ); } else { seen = 1; }
        data += sprintf( data, "Enter" );
    }
    if( mask & PointerMotionMask || mask & PointerMotionHintMask ||
        mask & Button1MotionMask || mask & Button2MotionMask ||
        mask & Button3MotionMask || mask & Button4MotionMask ||
        mask & Button5MotionMask || mask & ButtonMotionMask ) {
        if( seen ) { data += sprintf( data, ", " ); } else { seen = 1; }
        data += sprintf( data, "Mouse" );
    }
    if( mask & ExposureMask ) {
        if( seen ) { data += sprintf( data, ", " ); } else { seen = 1; }
        data += sprintf( data, "Expose" );
    }
    if( mask & VisibilityChangeMask ) {
        if( seen ) { data += sprintf( data, ", " ); } else { seen = 1; }
        data += sprintf( data, "Visibility" );
    }
    if( mask & StructureNotifyMask || mask & ResizeRedirectMask ) {
        if( seen ) { data += sprintf( data, ", " ); } else { seen = 1; }
        data += sprintf( data, "Resize" );
    }
    if( mask & SubstructureNotifyMask || mask & SubstructureRedirectMask ) {
        if( seen ) { data += sprintf( data, ", " ); } else { seen = 1; }
        data += sprintf( data, "Substruct" );
    }
    if( mask & FocusChangeMask ) {
        if( seen ) { data += sprintf( data, ", " ); } else { seen = 1; }
        data += sprintf( data, "Focus" );
    }
    if( mask & ColormapChangeMask ) {
        if( seen ) { data += sprintf( data, ", " ); } else { seen = 1; }
        data += sprintf( data, "Colormap" );
    }
    if( mask & OwnerGrabButtonMask ) {
        if( seen ) { data += sprintf( data, ", " ); } else { seen = 1; }
        data += sprintf( data, "Grab" );
    }
}

void print_name( char *name, Display *dpy, Window win, Window root )
{
    char *wname;
    if( win == root ) {
        sprintf( name, "Root" );
    } else if( !XFetchName( dpy, win, &wname ) ) {
        sprintf( name, " " );
    } else {
        sprintf( name, "%s", wname );
    }
}


GtkTreeStore *store;

static gboolean check_events( gpointer data )
{
    Display *dpy = gdk_x11_get_default_xdisplay();
    Window real_root = gdk_x11_get_default_root_xwindow();
    Window win, root, child;
    int rx, ry;
    unsigned int mask;
    GtkTreeIter iter;

    gtk_tree_store_clear( store );
    gtk_tree_store_append( store, &iter, 0 );

    win = real_root;

    while( win ) {
        XWindowAttributes attrs;
        int x, y;
        char id[ 30 ];
        char name[ 30 ];
        char xtext[ 30 ];
        char ytext[ 30 ];
        char wtext[ 30 ];
        char htext[ 30 ];
        char override[ 30 ];
        char eventmask[ 255 ];

        XGetWindowAttributes( dpy, win, &attrs );
        sprintf( id, "%x", (unsigned int) win );
        sprintf( xtext, "%d", attrs.x );
        sprintf( ytext, "%d", attrs.y );
        sprintf( wtext, "%d", attrs.width );
        sprintf( htext, "%d", attrs.height );
        print_event_mask( eventmask, attrs.all_event_masks );
        print_name( name, dpy, win, real_root );
        sprintf( override, "%s", attrs.override_redirect ? "Override" : "" );

        gtk_tree_store_set( store, &iter,
                            ID_COLUMN, id,
                            NAME_COLUMN, name,
                            X_COLUMN, xtext,
                            Y_COLUMN, ytext,
                            W_COLUMN, wtext,
                            H_COLUMN, htext,
                            EVENTS_COLUMN, eventmask,
                            OVERRIDE_COLUMN, override, -1 );
        gtk_tree_store_append( store, &iter, 0 );
        XQueryPointer(dpy, win, &root, &child, &rx, &ry, &x, &y, &mask );
        win = child;
    }

    return TRUE;
}

int main( int argc, char **argv )
{
    GtkWidget *window;
    GtkWidget *vbox;
    GtkWidget *tree;
    GtkCellRenderer *renderer;
    GtkTreeViewColumn *column;

    gtk_init( &argc, &argv );

    window = gtk_window_new( GTK_WINDOW_TOPLEVEL );

    g_signal_connect( G_OBJECT( window ), "destroy",
                      G_CALLBACK( on_destroy ), 0 );

    vbox = gtk_vbox_new( FALSE, 2 );
    gtk_container_add( GTK_CONTAINER( window ), vbox );

    store = gtk_tree_store_new( N_COLUMNS,
                                G_TYPE_STRING,  // id
                                G_TYPE_STRING,  // name
                                G_TYPE_STRING,  // x
                                G_TYPE_STRING,  // y
                                G_TYPE_STRING,  // w
                                G_TYPE_STRING,  // h
                                G_TYPE_STRING,  // events
                                G_TYPE_STRING); // override
    tree = gtk_tree_view_new_with_model( GTK_TREE_MODEL( store ) );

    renderer = gtk_cell_renderer_text_new();
    column = gtk_tree_view_column_new_with_attributes( "Window", renderer, "text", ID_COLUMN, 0 );
    gtk_tree_view_append_column( GTK_TREE_VIEW( tree ), column );

    renderer = gtk_cell_renderer_text_new();
    column = gtk_tree_view_column_new_with_attributes( "Name", renderer, "text", NAME_COLUMN, 0 );
    gtk_tree_view_column_set_max_width( column, 100 );
    gtk_tree_view_append_column( GTK_TREE_VIEW( tree ), column );

    renderer = gtk_cell_renderer_text_new();
    column = gtk_tree_view_column_new_with_attributes( "X", renderer, "text", X_COLUMN, 0 );
    gtk_tree_view_append_column( GTK_TREE_VIEW( tree ), column );

    renderer = gtk_cell_renderer_text_new();
    column = gtk_tree_view_column_new_with_attributes( "Y", renderer, "text", Y_COLUMN, 0 );
    gtk_tree_view_append_column( GTK_TREE_VIEW( tree ), column );
    renderer = gtk_cell_renderer_text_new();
    column = gtk_tree_view_column_new_with_attributes( "Width", renderer, "text", W_COLUMN, 0 );
    gtk_tree_view_append_column( GTK_TREE_VIEW( tree ), column );
    renderer = gtk_cell_renderer_text_new();
    column = gtk_tree_view_column_new_with_attributes( "Height", renderer, "text", H_COLUMN, 0 );
    gtk_tree_view_append_column( GTK_TREE_VIEW( tree ), column );

    renderer = gtk_cell_renderer_text_new();
    column = gtk_tree_view_column_new_with_attributes( "Events", renderer, "text", EVENTS_COLUMN, 0 );
    gtk_tree_view_column_set_max_width( column, 100 );
    gtk_tree_view_append_column( GTK_TREE_VIEW( tree ), column );

    renderer = gtk_cell_renderer_text_new();
    column = gtk_tree_view_column_new_with_attributes( "Override", renderer, "text", OVERRIDE_COLUMN, 0 );
    gtk_tree_view_append_column( GTK_TREE_VIEW( tree ), column );

    gtk_box_pack_start( GTK_BOX( vbox ), tree, TRUE, TRUE, 0 );

    g_timeout_add( 50, check_events, window );

    gtk_widget_show_all( window );

    gtk_main();
    return 0;
}

